'use strict';

/**
 * 静态文件路径 (404.html, sw.js, conf.js)
 */
const ASSET_URL = 'https://page.taw.us.kg';

const JS_VER = 10;
const MAX_RETRY = 3;

/** @type {ResponseInit} */
const PREFLIGHT_INIT = {
  status: 204,
  headers: new Headers({
    'access-control-allow-origin': '*',
    'access-control-allow-methods':
      'GET, POST, PUT, PATCH, TRACE, DELETE, HEAD, OPTIONS',
    'access-control-max-age': '1728000',
  }),
};

/**
 * 创建响应
 * @param {any} body
 * @param {number} [status=200]
 * @param {Object<string, string>} [headers={}]
 */
function makeRes(body, status = 200, headers = {}) {
  headers['--ver'] = JS_VER;
  headers['access-control-allow-origin'] = '*';
  return new Response(body, { status, headers });
}

/**
 * 解析 URL 字符串
 * @param {string} urlStr
 * @returns {URL|null}
 */
function newUrl(urlStr) {
  try {
    return new URL(urlStr);
  } catch (err) {
    return null;
  }
}

addEventListener('fetch', (e) => {
  const ret = fetchHandler(e).catch((err) =>
    makeRes('cfworker error:\n' + err.stack, 502)
  );
  e.respondWith(ret);
});

/**
 * 处理请求事件
 * @param {FetchEvent} e
 */
async function fetchHandler(e) {
  const req = e.request;
  const urlStr = req.url;
  const urlObj = new URL(urlStr);
  const path = urlObj.href.substr(urlObj.origin.length);

  if (urlObj.protocol === 'http:') {
    urlObj.protocol = 'https:';
    return makeRes('', 301, {
      'strict-transport-security': 'max-age=99999999; includeSubDomains; preload',
      location: urlObj.href,
    });
  }

  if (path.startsWith('/http/')) {
    return httpHandler(req, path.substr(6));
  }

  if (path === '/ws') {
    return handleWebSocket(req);
  }
  
  if (path === '/sw.js') {
    return fetch(ASSET_URL + '/sw.js');
  }
  // 添加登录路径处理
  if (path === '/login') {
    return loginHandler(req);
  }

  switch (path) {
    case '/http':
      return makeRes('请更新 cfworker 到最新版本!');
    case '/works':
      return makeRes('it works');
    default:
      // 静态文件
      return fetch(ASSET_URL + path);
  }
}

/**
 * 处理 WebSocket 请求
 * @param {Request} req
 */
async function handleWebSocket(req) {
  const upgradeHeader = req.headers.get('Upgrade');

  if (upgradeHeader && upgradeHeader.toLowerCase() === 'websocket') {
    // 处理 WebSocket 升级
    return WebSocket.accept(req);
  } else {
    return makeRes('WebSocket upgrade required', 400);
  }
}

/**
 * 处理登录请求
 * @param {Request} req
 */
async function loginHandler(req) {
  if (req.method !== 'POST') {
    return makeRes('Method Not Allowed', 405);
  }

  const body = await req.json();
  const { loginUrl, username, password } = body;

  // 创建登录请求
  const loginResponse = await fetch(loginUrl, {
    method: 'POST',
    headers: {
      'Content-Type': 'application/json', // 根据目标网站的要求调整
    },
    body: JSON.stringify({ username, password }), // 根据目标网站的登录要求调整
  });

  if (loginResponse.ok) {
    const data = await loginResponse.json();
    return makeRes(JSON.stringify(data), 200, {
      'Content-Type': 'application/json',
    });
  } else {
    return makeRes('Login failed', loginResponse.status);
  }
}

/**
 * 处理 HTTP 代理请求
 * @param {Request} req
 * @param {string} pathname
 */
async function httpHandler(req, pathname) {
  const reqHdrRaw = req.headers;
  if (reqHdrRaw.has('x-jsproxy')) {
    return Response.error();
  }

  // 预检请求
  if (
    req.method === 'OPTIONS' &&
    reqHdrRaw.has('access-control-request-headers')
  ) {
    return new Response(null, PREFLIGHT_INIT);
  }

  let rawSvr = '';
  let rawLen = '';
  let rawEtag = '';
  let acehOld = false;

  const reqHdrNew = new Headers(reqHdrRaw);
  reqHdrNew.set('x-jsproxy', '1');

  // 解析 referer 中的参数
  const refer = reqHdrNew.get('referer') || '';
  const queryIndex = refer.indexOf('?');
  if (queryIndex === -1) {
    return makeRes('missing params', 403);
  }
  const query = refer.substr(queryIndex + 1);
  const param = new URLSearchParams(query);

  for (const [k, v] of param.entries()) {
    if (k.startsWith('--')) {
      // 系统信息
      switch (k.substr(2)) {
        case 'aceh':
          acehOld = true;
          break;
        case 'raw-info':
          [rawSvr, rawLen, rawEtag] = v.split('|');
          break;
      }
    } else {
      // 还原 HTTP 请求头
      if (v) {
        reqHdrNew.set(k, v);
      } else {
        reqHdrNew.delete(k);
      }
    }
  }
  if (!param.has('referer')) {
    reqHdrNew.delete('referer');
  }

  // 修复路径中的 `//` 被合并成 `/` 的问题
  const urlStr = pathname.replace(/^(https?):\/+/, '$1://').replace(/-----/, '');
  const urlObj = newUrl(urlStr);
  if (!urlObj) {
    return makeRes('invalid proxy url: ' + urlStr, 403);
  }

  /** @type {RequestInit} */
  const reqInit = {
    method: req.method,
    headers: reqHdrNew,
    redirect: 'manual',
  };
  if (req.method === 'POST') {
    reqInit.body = req.body;
  }
  return proxy(urlObj, reqInit, acehOld, rawLen, 0);
}

/**
 * 代理请求并处理响应
 * @param {URL} urlObj
 * @param {RequestInit} reqInit
 * @param {boolean} acehOld
 * @param {string} rawLen
 * @param {number} retryTimes
 */
async function proxy(urlObj, reqInit, acehOld, rawLen, retryTimes) {
  const res = await fetch(urlObj.href, reqInit);
  const resHdrOld = res.headers;
  const resHdrNew = new Headers(resHdrOld);

  let expose = '*';

  for (const [k, v] of resHdrOld.entries()) {
    if (
      k === 'access-control-allow-origin' ||
      k === 'access-control-expose-headers' ||
      k === 'location' ||
      k === 'set-cookie'
    ) {
      const x = '--' + k;
      resHdrNew.set(x, v);
      if (acehOld) {
        expose = expose + ',' + x;
      }
      resHdrNew.delete(k);
    } else if (
      acehOld &&
      k !== 'cache-control' &&
      k !== 'content-language' &&
      k !== 'content-type' &&
      k !== 'expires' &&
      k !== 'last-modified' &&
      k !== 'pragma'
    ) {
      expose = expose + ',' + k;
    }
  }

  if (acehOld) {
    expose = expose + ',--s';
    resHdrNew.set('--t', '1');
  }

  // 验证内容长度
  if (rawLen) {
    const newLen = resHdrOld.get('content-length') || '';
    const badLen = rawLen !== newLen;

    if (badLen) {
      if (retryTimes < MAX_RETRY) {
        const newUrlObj = await parseYtVideoRedir(urlObj, newLen, res);
        if (newUrlObj) {
          return proxy(newUrlObj, reqInit, acehOld, rawLen, retryTimes + 1);
        }
      }
      return makeRes(res.body, 400, {
        '--error': `bad len: ${newLen}, expected: ${rawLen}`,
        'access-control-expose-headers': '--error',
      });
    }

    if (retryTimes > 1) {
      resHdrNew.set('--retry', retryTimes);
    }
  }

  let status = res.status;

  resHdrNew.set('access-control-expose-headers', expose);
  resHdrNew.set('access-control-allow-origin', '*');
  resHdrNew.set('--s', status);
  resHdrNew.set('--ver', JS_VER);

  resHdrNew.delete('content-security-policy');
  resHdrNew.delete('content-security-policy-report-only');
  resHdrNew.delete('clear-site-data');

  if ([301, 302, 303, 307, 308].includes(status)) {
    status += 10; // 修改状态码以避免循环重定向
  }
  return makeRes(res.body, status, resHdrNew);
}

/**
 * 解析 YouTube 视频重定向
 * @param {URL} urlObj
 * @param {string} rawLen
 * @param {Response} res
 */
async function parseYtVideoRedir(urlObj, rawLen, res) {
  const newUrlStr = res.headers.get('location') || '';

  if (!newUrlStr) {
    return null;
  }

  const newUrlObj = newUrl(newUrlStr);
  if (!newUrlObj) {
    return null;
  }

  const newPath = newUrlObj.pathname;

  if (
    newPath.startsWith('/get_video_info') ||
    newPath.startsWith('/v/') ||
    newPath.startsWith('/watch')
  ) {
    return newUrlObj;
  }
  return null;
}
